@@ -24,7 +24,13 @@ import (
 var trans = make(map[string]string)
 var attrRe = regexp.MustCompile(`\{\{\s*'([^']+)'\s+\|\s+translate\s*\}\}`)
 var attrReCond = regexp.MustCompile(`\{\{.+\s+\?\s+'([^']+)'\s+:\s+'([^']+)'\s+\|\s+translate\s*\}\}`)
-var jsRe = regexp.MustCompile(`\$translate.instant\("([^"]+)"\)`)
+
+// Find both $translate.instant("…") and $translate.instant("…",…) in JS.
+// Consider single quote variants too.
+var jsRe = []*regexp.Regexp{
+	regexp.MustCompile(`\$translate\.instant\(\s*"(.+?)"(,.*|\s*)\)`),
+	regexp.MustCompile(`\$translate\.instant\(\s*'(.+?)'(,.*|\s*)\)`),
+}
 
 // exceptions to the untranslated text warning
 var noStringRe = regexp.MustCompile(
@@ -128,8 +134,10 @@ func walkerFor(basePath string) filepath.WalkFunc {
 			generalNode(doc, filepath.Base(name))
 		case ".js":
 			for s := bufio.NewScanner(fd); s.Scan(); {
-				for _, matches := range jsRe.FindAllStringSubmatch(s.Text(), -1) {
-					translation(matches[1])
+				for _, re := range jsRe {
+					for _, matches := range re.FindAllStringSubmatch(s.Text(), -1) {
+						translation(matches[1])
+					}
 				}
 			}
 		}
@@ -138,6 +146,22 @@ func walkerFor(basePath string) filepath.WalkFunc {
 	}
 }
 
+func collectThemes(basePath string) {
+	files, err := os.ReadDir(basePath)
+	if err != nil {
+		log.Fatal(err)
+	}
+	for _, f := range files {
+		if f.IsDir() {
+			key := "theme-name-" + f.Name()
+			if _, ok := trans[key]; !ok {
+				name := strings.Title(f.Name())
+				trans[key] = name
+			}
+		}
+	}
+}
+
 func main() {
 	fd, err := os.Open(os.Args[1])
 	if err != nil {
@@ -152,6 +176,7 @@ func main() {
 	var guiDir = os.Args[2]
 
 	filepath.Walk(guiDir, walkerFor(guiDir))
+	collectThemes(guiDir)
 
 	bs, err := json.MarshalIndent(trans, "", "   ")
 	if err != nil {
